import random
import time

import pygame
import pgu.gui
import pgu.html

import camera
import config
import image_cache
import person
import general
import screen
import high_scores
import music

class TileTypes:
	Ground		= "ground"
	Grass		= "grass"
	Flowers		= "flowers"
	Concrete	= "concrete"
	Plaza		= "plaza"

class Level(screen.Screen):
	def __init__(self, level_config, pgu_container):
		screen.Screen.__init__(self,pgu_container)
		self.start_time = None
		self.level_config = level_config
		self.camera = camera.Camera(self.level_config.size)
		self.tile_map = [[TileTypes.Grass,0.0]]*(self.level_config.size[0]*self.level_config.size[1])
		self.last_mouse_position = None
		self.time_since_last_person = 0
		self.win_time = None
		
		self.people = []
		self.money = self.level_config.start_money
		self.gayness = 0
		
		#people label
		pgu_container.add(pgu.gui.basic.Image(image_cache.get_cache_image("man_small",True)),7,5)
		self.label_people_on_grass = pgu.gui.basic.Label("************")
		pgu_container.add(self.label_people_on_grass,30,10)
		
		#money label
		pgu_container.add(pgu.gui.basic.Image(image_cache.get_cache_image("money",True)),2,38)
		self.label_money = pgu.gui.basic.Label("*************",width=500,align = -1)
		pgu_container.add(self.label_money,30,40)
		
		#time label
		pgu_container.add(pgu.gui.basic.Image(image_cache.get_cache_image("time",True)),4,65)
		self.label_time = pgu.gui.basic.Label("*************",width=500,align = -1)
		pgu_container.add(self.label_time,30,67)
		
		#gayness label		
		self.label_gayness = pgu.gui.basic.Label("*"*50)
		pgu_container.add(self.label_gayness,10,90)
		
		#gay bar
		self.gayness_bar = pgu.gui.misc.ProgressBar(250,0,1000,width = 300)
		pgu_container.add(self.gayness_bar,(config.screen_size[0]-800)/2 + 250,config.screen_size[1]-45)
		
		#tools
		tool_bar = pgu.gui.document.Document()
		self.tool_group = pgu.gui.Group(name='colors',value=TileTypes.Grass)
		def add_tool(image_name, value):
			image = pgu.gui.basic.Image(image_cache.get_cache_image(image_name,True))
			tool = pgu.gui.Tool(self.tool_group, image, value ,height=40,width = 40)
			return tool
		tool_bar.add(add_tool("grass_button",TileTypes.Grass))
		if self.level_config.flowers_enabled:
			tool_bar.space((10,0))
			tool_bar.add(add_tool("flowers_button",TileTypes.Flowers))
		if self.level_config.concrete_enabled:
			tool_bar.space((10,0))
			tool_bar.add(add_tool("concrete_button",TileTypes.Concrete))
		if self.level_config.plaza_enabled:
			tool_bar.space((10,0))
			tool_bar.add(add_tool("plaza_button",TileTypes.Plaza))
		pgu_container.add(tool_bar,5,config.screen_size[1]-45)
		
		#'restart' and 'main menu' buttons:
		restart_button=pgu.gui.button.Button("Restart")
		restart_button.connect(pgu.gui.CLICK,self.restart)
		main_menu_button=pgu.gui.button.Button("Main menu")
		main_menu_button.connect(pgu.gui.CLICK,self.go_to_main_menu)
		
		buttons_table = pgu.gui.table.Table()
		buttons_table.tr()
		buttons_table.td(restart_button, align=0)
		buttons_table.tr()
		buttons_table.td(pgu.gui.basic.Spacer(1,3), align=0)
		buttons_table.tr()
		buttons_table.td(main_menu_button, align=0)
		pgu_container.add(buttons_table,config.screen_size[0]-108,config.screen_size[1]-47)
		
		self.opening_dialog_opened = False
	
	def start_game(self):
		self.start_time = time.time()
	
	def restart(self):
		self.done = True
		level_class = self.level_config
		try:
			level_class = level_class.__class__
		except:
			pass
		self.next_screen = config.all_levels.index(level_class)
		
	def create_winning_dialog(self):
		self.win_time = time.time()
		
		#gui:
		continue_button=pgu.gui.button.Button("Continue")
		
		this_time = self.win_time - self.start_time
		finish_label = pgu.gui.basic.Label("You finished the level in %.2f seconds!" % this_time)
		
		best_so_far = high_scores.check(self.level_config)
		if best_so_far == None or this_time < best_so_far:
			high_scores.update(self.level_config,this_time)
			best_text = "You have beaten the high score for this level!"
		else:
			best_text = "The best score for this level is %.2f seconds!" % best_so_far
		best_label = pgu.gui.basic.Label(best_text)
		
		dialog_table = pgu.gui.table.Table()
		dialog_table.tr()
		dialog_table.td(finish_label, align=0)
		dialog_table.tr()
		dialog_table.td(pgu.gui.basic.Spacer(1,10), align=0)
		dialog_table.tr()
		dialog_table.td(best_label, align=0)
		dialog_table.tr()
		dialog_table.td(pgu.gui.basic.Spacer(1,10), align=0)
		dialog_table.tr()
		dialog_table.td(continue_button, align=0)
		dialog_table.tr()
		dialog_table.td(pgu.gui.basic.Spacer(1,4), align=0)
		
		title = pgu.gui.basic.Label("You won!")
		winning_dialog = pgu.gui.dialog.Dialog(title,dialog_table)
		winning_dialog.connect(pgu.gui.CLOSE,self.win_ok_pressed)
		continue_button.connect(pgu.gui.CLICK,winning_dialog.close)
		winning_dialog.open()

	def win_ok_pressed(self):
		level_class = self.level_config
		try:
			level_class = level_class.__class__
		except:
			pass
		next_level_index = config.all_levels.index(level_class) + 1
		if len(config.all_levels) == next_level_index:
			next_level_index = "win"
		self.next_screen = next_level_index
		self.done = True
	
	def draw_tile_at(self, target, tile, position, opacity = None):
		screen_position = self.camera.translate_tile_to_screen(position)
		
		#we point to the left corner of the tile, while drawing at the corner
		screen_position = (screen_position.x,screen_position.y-config.tile_size[1]/2)
		
		#basic clipping
		if screen_position[0] > config.screen_size[0]:
			return
		if screen_position[0] < -config.tile_size[0]:
			return
		if screen_position[1] > config.screen_size[1]:
			return
		if screen_position[1] < -config.tile_size[1]:
			return
		
		if opacity == None:
			target.blit(tile,screen_position)
		else:
			tile.set_alpha(opacity*255)
			target.blit(tile,screen_position)
	
	def mouse_move(self, position):
		self.last_mouse_position = position
		self.camera.mouse_move(position)
	
	def key_down(self, key):
		if self.start_time == None:
			self.opening_dialog.close()
		
		self.camera.key_down(key)
		
		def set_tool(type):
			self.tool_group.value = type
			self.refresh_gui()
		if key == pygame.K_1:
			set_tool(TileTypes.Grass)
		elif key == pygame.K_2:
			set_tool(TileTypes.Flowers)
		elif key == pygame.K_3:
			set_tool(TileTypes.Concrete)
		elif key == pygame.K_4:
			set_tool(TileTypes.Plaza)
		
		if key == pygame.K_s and False:
			self.win_ok_pressed()
		if key == pygame.K_m:
			music.switch_music()
	def key_up(self, key):
		self.camera.key_up(key)
	
	def out_of_screen(self):
		self.last_mouse_position = None
		self.camera.out_of_screen()
	
	def in_game(self):
		return self.start_time != None and self.win_time == None
	def mouse_click(self, position):
		if self.last_mouse_position == None:
			return #out of level area
		
		if not self.in_game():
			return
		
		position = self.camera.translate_screen_to_tile(position)
		if not self.is_tile_in_level_area(position):
			return
		
		index = self.get_tile_index(position.x,position.y)
		if self.tool_group.value == TileTypes.Grass:
			cost = self.level_config.grass_cost
		elif self.tool_group.value == TileTypes.Flowers:
			cost = self.level_config.flowers_cost
		elif self.tool_group.value == TileTypes.Concrete:
			cost = self.level_config.concrete_cost
		elif self.tool_group.value == TileTypes.Plaza:
			cost = self.level_config.plaza_cost
		if self.money < cost:
			return
		
		if self.tile_map[index] == [self.tool_group.value,1.0]:
			return
		
		self.tile_map[index] = [self.tool_group.value,1.0]
		self.money -= cost
	
	def get_tile_index(self, x, y):
		return int(x)+int(y)*self.level_config.size[0]
	
	def create_opening_dialog(self):
		html = pgu.html.HTML(self.level_config.html)
		start_button = pgu.gui.button.Button(pgu.gui.basic.Label("Start"))
		title = pgu.gui.basic.Label(self.level_config.description)
		
		dialog_table = pgu.gui.table.Table()
		dialog_table.tr()
		dialog_table.td(html, align=0)
		dialog_table.tr()
		dialog_table.td(pgu.gui.basic.Spacer(1,5), align=0)
		dialog_table.tr()
		dialog_table.td(start_button, align=0)
		
		self.opening_dialog = pgu.gui.dialog.Dialog(title,dialog_table)
		self.opening_dialog.connect(pgu.gui.CLOSE,self.start_game)
		start_button.connect(pgu.gui.CLICK,self.opening_dialog.close)
		self.opening_dialog.open()
		self.opening_dialog_opened = True
	
	def get_gayness_rate(self):
		rate = self.gayness / self.level_config.gayness_target
		if rate <0.0:
			rate = 0.0
		if rate >1.0:
			rate = 1.0
		return rate
	
	def do_frame(self, time_delta):
		if self.in_game():
			self.camera.do_frame(time_delta)
		
		if not self.opening_dialog_opened:
			self.create_opening_dialog()
		
		if not self.in_game():
			return
		
		for person_instance in self.people:
			person_instance.do_frame(time_delta)
		
		#remove people that are gone
		def person_in_level(person):
			if person.position.x < 0 or person.position.y < 0:
				return False
			if int(person.position.x) >= self.level_config.size[0] or int(person.position.y) >= self.level_config.size[1]:
				return False
			return True
		self.people = filter(person_in_level, self.people)
		
		#degrade the grass
		for person_instance in self.people:
			tile_index = self.get_tile_index(person_instance.position.x,person_instance.position.y)
			if self.tile_map[tile_index][0] == TileTypes.Grass or self.tile_map[tile_index][0] == TileTypes.Flowers:
				self.tile_map[tile_index][1] -= time_delta * self.level_config.degradation_factor
				if self.tile_map[tile_index][1] < 0:
					self.tile_map[tile_index][1] = 0
		
		#calculate gayness
		gayness = 0.0
		for tile_val in self.tile_map:
			if tile_val[0] == TileTypes.Grass:
				gayness += tile_val[1] * self.level_config.grass_gayness
			elif tile_val[0] == TileTypes.Flowers:
				gayness += self.level_config.flowers_min_gayness + tile_val[1] * (self.level_config.flowers_max_gayness - self.level_config.flowers_min_gayness)
			elif tile_val[0] == TileTypes.Concrete:
				gayness += self.level_config.concrete_gayness
			elif tile_val[0] == TileTypes.Plaza:
				gayness += self.level_config.plaza_gayness
		self.gayness = gayness
		
		#spawn people
		self.time_since_last_person += time_delta
		rate = self.level_config.minimum_person_rate + (self.level_config.maximum_person_rate - self.level_config.minimum_person_rate) * self.get_gayness_rate()
		time_to_wait = 1/rate
		if self.time_since_last_person > time_to_wait:
			self.time_since_last_person = 0
			self.money += self.level_config.money_per_visitor
			self.people.append(person.Person(self.level_config))
		
		#win
		if self.gayness >= self.level_config.gayness_target:
			self.create_winning_dialog()
	
	def is_tile_in_level_area(self,position):
		return position.x >= 0 and position.x < self.level_config.size[0] and position.y >= 0 and position.y < self.level_config.size[1]
	
	def draw(self,target):
		#draw the tiles
		for i in range(self.level_config.size[0]):
			for j in range(self.level_config.size[1]):
				tile = self.tile_map[self.get_tile_index(i,j)]
				if tile[0] == TileTypes.Grass:
					self.draw_tile_at(target, image_cache.get_cache_tile("ground"), general.Vector2(i,j) )
					self.draw_tile_at(target, image_cache.get_cache_tile(TileTypes.Grass), general.Vector2(i,j), tile[1])
				elif tile[0] == TileTypes.Flowers:
					self.draw_tile_at(target, image_cache.get_cache_tile("dead_flowers"), general.Vector2(i,j) )
					self.draw_tile_at(target, image_cache.get_cache_tile(TileTypes.Flowers), general.Vector2(i,j), tile[1])
				elif tile[0] == TileTypes.Concrete:
					self.draw_tile_at(target, image_cache.get_cache_tile(TileTypes.Concrete), general.Vector2(i,j) )
				elif tile[0] == TileTypes.Plaza:
					self.draw_tile_at(target, image_cache.get_cache_tile(TileTypes.Plaza), general.Vector2(i,j) )
		
		#mouse highlighting:
		if self.in_game():
			if self.last_mouse_position != None:
				position = self.camera.translate_screen_to_tile(self.last_mouse_position)
				if self.is_tile_in_level_area(position):
					position = general.Vector2(int(position.x),int(position.y))
					self.draw_tile_at(target, image_cache.get_cache_tile("highlight",True), position)
		
		#draw the people
		people_to_draw = []
		for person_instance in self.people:
			pos_on_screen = self.camera.translate_tile_to_screen(person_instance.position)
			people_to_draw.append((person_instance, pos_on_screen))
		people_to_draw.sort(cmp = lambda x,y: cmp(x[1].y,y[1].y))
		for person_instance, pos_on_screen in people_to_draw:
			person_instance.draw(target, pos_on_screen)
		
		#update the labels
		self.label_people_on_grass.value = str(len(self.people))
		self.label_gayness.value = "Lovelyness: %.2f / %.2f" % (self.gayness, self.level_config.gayness_target)
		self.label_money.value = "%.2f" % self.money
		self.gayness_bar.value = int(1000 * self.get_gayness_rate())
		if self.start_time == None:
			self.label_time.value = "0"
		else:
			if self.win_time == None:
				time_passed = time.time() - self.start_time
			else:
				time_passed = self.win_time - self.start_time
			time_passed = int(time_passed)
			if time_passed>=60:
				self.label_time.value = "%d:%02d" % (time_passed/60,time_passed%60)
			else:
				self.label_time.value = str(time_passed)
		
		#draw the game area (bottom line only)
		pygame.draw.line(target, (0,0,0), (config.game_area[0],config.game_area[3]), (config.game_area[2],config.game_area[3]), 1)
